IMPLEMENTATION MODULE Lists;

(*
 * Jede Liste besitzt einen Cursor, der den aktuellen Eintrag bestimmt,
 * sowie einen Wurzeleintrag, Root genannt.
 * Eine Liste ist als Ring angelegt, wobei der Root-Eintrag vor dem ersten
 * und hinter dem letzten echten Eintrag liegt.
 *)


FROM     SYSTEM         IMPORT   ADDRESS;
FROM     System         IMPORT   Move;
FROM     Storage        IMPORT   ALLOCATE, DEALLOCATE;
IMPORT   ModReg;
IMPORT   void;

(*
TYPE     List;

         LDir        = (lForward,    (* vorwrts  *)
                        lBackward);  (* rckwrts *)

         LCopyMode   = (lFlat,       (* die vorhandenen Zeiger werden bernommen *)
                        lDeep);      (* es werden Kopien der Eintrge angelegt *)

         LDeleteEntryProc  =  PROCEDURE((*entry:*) ADDRESS);
         
         LCopyEntryProc    =  PROCEDURE((*entry:*) ADDRESS): ADDRESS;

         LCondProc         =  PROCEDURE((*entry:*) ADDRESS,
                                        (*parm: *) ADDRESS): BOOLEAN;

         LDoProc           =  PROCEDURE((*entry:*) ADDRESS,
                                        (*parm: *) ADDRESS);
*)

TYPE     PtrNode     =  POINTER TO ListNode;

         ListNode    =  RECORD
                           data:       ADDRESS;
                           prev,
                           next:       PtrNode;
                        END;

         ListHead    =  RECORD
                           root,
                           cursor:     PtrNode;
                           entries:    LONGCARD;
                           delProc:    LDeleteEntryProc;
                           copyProc:   LCopyEntryProc;
                        END;

         List        =  POINTER TO ListHead;


VAR   error:   BOOLEAN;
      

(*$H+*)

PROCEDURE Create(VAR list:     List;
                     delProc:  LDeleteEntryProc;
                     copyProc: LCopyEntryProc): BOOLEAN;
   (*
    * TRUE, wenn eine neue Liste angelegt werden konnte,
    * FALSE sonst.
    * Bei einer neuen Liste zeigt Cursor auf Root.
    *)
VAR   result:  BOOLEAN;
      root:    PtrNode;
BEGIN
   result := FALSE;
   NEW(root);
   IF root#NIL THEN
      NEW(list);
      IF list#NIL THEN
         result := TRUE;
         root^.data := NIL;
         root^.prev := root;
         root^.next := root;
         list^.root := root;
         list^.cursor := root;
         list^.entries := 0L;
         list^.delProc := delProc;
         list^.copyProc := copyProc;
      ELSE
         DISPOSE(root);
      END;
   END;
   RETURN result;
END Create;

PROCEDURE Inherit(VAR source,
                      dest:   List): BOOLEAN;
   (*
    * Legt eine neue Liste an, die die gleichen Delete-
    * und Copy-Prozeduren hat wie source.
    * TRUE, wenn eine neue Liste angelegt werden konnte,
    * FALSE sonst.
    * Bei einer neuen Liste zeigt Cursor auf Root.
    *)
BEGIN
   RETURN Create(dest, source^.delProc, source^.copyProc);
END Inherit;
   
PROCEDURE Delete(VAR list: List): BOOLEAN;
   (*
    * Lscht eine Liste samt aller Eintrge.
    * TRUE, wenn die Liste komplett gelscht werden konnte,
    * FALSE sonst. Im Fehlerfall hat list den Wert NoList.
    *)
VAR   entry:   ADDRESS;
      node:    PtrNode;
BEGIN
   node := list^.root^.next;
   WHILE (node#list^.root) DO
      list^.delProc(node^.data);
      node := node^.next;
      DISPOSE(node^.prev);
   END;
   DISPOSE(list^.root);
   DISPOSE(list);
   list := NIL;
   RETURN TRUE;
END Delete;

PROCEDURE Empty(list: List): BOOLEAN;
   (*
    * TRUE genau dann, wenn die Liste keine Eintrge (auer Root) enthlt.
    *)
BEGIN
   RETURN (list^.root^.next=list^.root);
END Empty;

PROCEDURE Entries(list: List): LONGCARD;
   (*
    * Liefert die Anzahl der Eintrge.
    *)
BEGIN
   RETURN list^.entries;
END Entries;



PROCEDURE InsertEntry(VAR list:  List;
                          entry: ADDRESS): BOOLEAN;
   (*
    * TRUE genau dann, wenn das Eintragen erfolgreich war.
    * Der Cursor wird auf entry gesetzt.
    *)
VAR   result:  BOOLEAN;
      node:    PtrNode;
BEGIN
   result := FALSE;
   NEW(node);
   IF node#NIL THEN
      node^.data := entry;
      node^.prev := list^.cursor;
      node^.next := list^.cursor^.next;
      node^.prev^.next := node;
      node^.next^.prev := node;
      list^.cursor := node;
      INC(list^.entries);
      result := TRUE;
   END;
   RETURN result;
END InsertEntry;

PROCEDURE RemoveEntry(VAR list:  List): BOOLEAN;
   (*
    * TRUE genau dann, wenn das Lschen erfolgreich war.
    * Der Cursor wird auf den Vorgnger von entry positioniert.
    *
    * ACHTUNG: Der Eintrag wird aus der Liste gelscht, aber nicht
    *          per list^.delProc freigegeben!
    *)
VAR   result:  BOOLEAN;
      node:    PtrNode;
BEGIN
   result := FALSE;
   IF list^.cursor#list^.root THEN
      node := list^.cursor;
      list^.delProc(node^.data);
      node^.prev^.next := node^.next;
      node^.next^.prev := node^.prev;
      list^.cursor := node^.prev;
      DISPOSE(node);
      DEC(list^.entries);
      result := TRUE;
   END;
   RETURN result;
END RemoveEntry;

PROCEDURE AppendEntry(VAR list:  List;
                          entry: ADDRESS): BOOLEAN;
   (*
    * Fgt entry am Ende der Liste an.
    * TRUE genau dann, wenn das Einfgen erfolgreich war.
    * Der Cursor bleibt unverndert.
    *)
VAR   result:  BOOLEAN;
      node:    PtrNode;
BEGIN
   result := FALSE;
   NEW(node);
   IF node#NIL THEN
      node^.data := entry;
      node^.prev := list^.root^.prev;
      node^.next := list^.root;
      node^.prev^.next := node;
      node^.next^.prev := node;
      INC(list^.entries);
      result := TRUE;
   END;
   RETURN result;
END AppendEntry;


PROCEDURE ResetCursor(VAR list: List);
   (*
    * Setzt den Cursor auf Root.
    *)
BEGIN
   list^.cursor := list^.root;
END ResetCursor;

PROCEDURE TestRoot(list: List): BOOLEAN;
   (*
    * liefert TRUE genau dann, wenn der Cursor auf Root zeigt.
    *)
BEGIN
   RETURN (list^.cursor=list^.root);
END TestRoot;


PROCEDURE TestFirstEntry(list: List): BOOLEAN;
   (*
    * liefert TRUE genau dann, wenn der Cursor auf den ersten Eintrag zeigt.
    *)
BEGIN
   RETURN (list^.cursor^.prev=list^.root);
END TestFirstEntry;

PROCEDURE TestLastEntry(list: List): BOOLEAN;
   (*
    * liefert TRUE genau dann, wenn der Cursor auf den letzten Eintrag zeigt.
    *)
BEGIN
   RETURN (list^.cursor^.next=list^.root);
END TestLastEntry;

PROCEDURE CurrentEntry(list: List): ADDRESS;
   (*
    * Liefert den aktuellen Eintrag.
    *)
BEGIN
   RETURN list^.cursor^.data;
END CurrentEntry;

PROCEDURE FirstEntry(VAR list: List): ADDRESS;
   (*
    * Liefert den ersten Eintrag. Root falls list leer.
    * Der Cursor wird entsprechend gesetzt.
    *)
VAR result: ADDRESS;
BEGIN
   list^.cursor := list^.root^.next;
   RETURN list^.cursor^.data;
END FirstEntry;

PROCEDURE LastEntry(VAR list: List): ADDRESS;
   (*
    * Liefert den letzten Eintrag. Root falls list leer.
    * Der Cursor wird entsprechend gesetzt.
    *)    
VAR result: ADDRESS;
BEGIN
   list^.cursor := list^.root^.prev;
   RETURN list^.cursor^.data;
END LastEntry;

PROCEDURE NextEntry(VAR list: List): ADDRESS;
   (*
    * Liefert den nchsten Eintrag, setzt den Cursor darauf.
    *)
BEGIN
   list^.cursor := list^.cursor^.next;
   RETURN list^.cursor^.data;
END NextEntry;

PROCEDURE PrevEntry(VAR list: List): ADDRESS;
   (*
    * Liefert den vorigen Eintrag, setzt den Cursor darauf.
    *)
BEGIN
   list^.cursor := list^.cursor^.prev;
   RETURN list^.cursor^.data;
END PrevEntry;

PROCEDURE FindEntry(VAR list:    List;
                        dir:     LDir;
                        entry:   ADDRESS): BOOLEAN;
   (*
    * Durchsucht list nach entry in Richtung dir.
    * Wird der Eintrag gefunden, zeigt der Cursor darauf.
    * Sonst zeigt der Cursor danach auf Root.
    * TRUE genau dann, wenn entry gefunden wurde.
    *)
VAR   result:  BOOLEAN;
      node:    PtrNode;
BEGIN
   result := FALSE;
   IF dir=lForward THEN
      node := list^.root^.next;
      WHILE NOT result AND (node#list^.root) DO
         result := (node^.data=entry);
         list^.cursor := node;
         node := node^.next;
      END;
   ELSE
      node := list^.root^.prev;
      WHILE NOT result AND (node#list^.root) DO
         result := (node^.data=entry);
         list^.cursor := node;
         node := node^.prev;
      END;
   END;
   IF NOT result THEN
      list^.cursor := list^.root;
   END;
   RETURN result;
END FindEntry;

PROCEDURE SearchEntry(VAR list:  List;
                          dir:   LDir;
                          cond:  LCondProc;
                          parm:  ADDRESS): ADDRESS;
   (*
    * Durchsucht list in Richtung dir. Liefert den ersten Eintrag, fr den
    * LCondProc TRUE ist. NIL, falls kein passender Eintrag vorhanden.
    * Der Cursor zeigt auf den zuletzt gefundenen Eintrag bzw. auf Root.
    * Die Suche beginnt _hinter_ bzw. _vor_ dem Cursor!
    *)                       
VAR   result:  ADDRESS;
      done:    BOOLEAN;
      node:    PtrNode;
BEGIN
   result := NIL;
   done := FALSE;
   IF dir=lForward THEN
      node := list^.cursor^.next;
      WHILE NOT done AND (node#list^.root) DO
         IF cond(node^.data, parm) THEN
            done := TRUE;
            list^.cursor := node;
            result := node^.data;
         END;
         node := node^.next;
      END;
   ELSE
      node := list^.cursor^.prev;
      WHILE NOT done AND (node#list^.root) DO
         IF cond(node^.data, parm) THEN
            done := TRUE;
            list^.cursor := node;
            result := node^.data;
         END;
         node := node^.prev;
      END;
   END;
   IF NOT done THEN
      list^.cursor := list^.root;
   END;
   RETURN result;
END SearchEntry;

PROCEDURE GetEntry(VAR list:  List;
                       dir:   LDir;
                       cond:  LCondProc;
                       parm:  ADDRESS): ADDRESS;
   (*
    * Wie SearchEntry, nur wird der Eintrag gleich aus der Liste gelscht.
    * Der Cursor zeigt auf Eintrag vor dem gelschten, bzw. auf Root.
    *
    * ACHTUNG: Der Eintrag wird aus der Liste gelscht, aber nicht
    *          per list^.delProc freigegeben!
    *)
VAR   result: ADDRESS;
BEGIN
   result := SearchEntry(list, dir, cond, parm);
   void.boolean := RemoveEntry(list);
   RETURN result;
END GetEntry;

PROCEDURE ScanList(VAR list:  List;
                       dir:   LDir;
                       cond:  LCondProc;
                       cParm: ADDRESS;
                       do:    LDoProc;
                       dParm: ADDRESS);
   (*
    * Durchluft die komplette Liste list in Richtung dir, und ruft fr jedes
    * Element, fr das cond TRUE liefert do auf.
    * Der Cursor wird nicht verndert.
    *)
VAR   node: PtrNode;
BEGIN
   IF dir=lForward THEN
      node := list^.root^.next;
      WHILE node#list^.root DO
         IF cond(node^.data, cParm) THEN
            do(node^.data, dParm);
         END;
         node := node^.next;
      END;
   ELSE
      node := list^.root^.prev;
      WHILE node#list^.root DO
         IF cond(node^.data, cParm) THEN
            do(node^.data, dParm);
         END;
         node := node^.prev;
      END;
   END;
END ScanList;

PROCEDURE CopyList(VAR source,
                       dest: List;
                       mode: LCopyMode): BOOLEAN;
   (*
    * Kopiert die Liste source auf dest. dest wird vorher gelscht,
    * falls ntig.
    * mode bestimmt, ob nur die Listenstruktur, oder auch die Eintrge
    * kopiert weden. Bei mode=lDeep werden die Eintrge mit der bei
    * Create angegebenen Prozedur kopiert.
    * Der Cursor zeigt danach bei beiden Listen auf Root.
    * TRUE, wenn das Kopieren geklappt hat, FALSE sonst.
    * Im Fehlerfall ist die Zielliste leer.
    *)
VAR  result:   BOOLEAN;
     node:     PtrNode;
BEGIN
   result := FALSE;
   IF Create(dest, source^.delProc, source^.copyProc) THEN
      error := FALSE;
      node := source^.root^.next;
      WHILE NOT error AND (node#source^.root) DO
         IF mode=lDeep THEN
            error := NOT AppendEntry(dest, source^.copyProc(node^.data));
         ELSE
            error := NOT AppendEntry(dest, node^.data);
         END;
         node := node^.next;
      END;
      IF error THEN
         void.boolean := Delete(dest);
      ELSE
         result := TRUE;
      END;
   END;
   RETURN result;
END CopyList;

PROCEDURE InsertList(VAR source,
                         dest:   List;
                         mode:   LCopyMode): BOOLEAN;
   (*
    * fgt die Elemente von list in to ein.
    * Der Cursor zeigt danach bei source auf Root und
    * bei dest auf den zuletzt eingefgten Eintrag.
    * TRUE, wenn das Kopieren geklappt hat, FALSE sonst.
    * Im Fehlerfall ist die Zielliste unverndert - hoffentlich ;-)
    *)
VAR   result:        BOOLEAN;
      error:         BOOLEAN;
      node:          PtrNode;
      beginInsert,
      endInsert:     PtrNode;
BEGIN
   result := FALSE;
   beginInsert := dest^.cursor;
   endInsert := dest^.cursor^.next;
   error := FALSE;
   node := source^.root^.next;
   WHILE (node#source^.root) AND ~error DO
      IF mode=lDeep THEN
         error := NOT InsertEntry(dest, source^.copyProc(node^.data));
      ELSE
         error := NOT InsertEntry(dest, node^.data);
      END;
   END;
   IF error THEN
      dest^.cursor := endInsert^.prev;
      WHILE dest^.cursor#beginInsert DO
         void.boolean := RemoveEntry(dest);
      END;
   ELSE
      result := TRUE;
   END;
   RETURN result;
END InsertList;


(* Standard-Proceduren fr LDeleteEntryProc *)

PROCEDURE DeleteNothing(entry: ADDRESS);
   (* Wie der Name schon sagt: Diese Prozedur tut gar nichts *)
BEGIN
END DeleteNothing;

(* Standard-Proceduren fr LCopyEntryProc *)

PROCEDURE CopyNothing(entry: ADDRESS): ADDRESS;
   (*
    * Wie der Name schon sagt: Diese Prozedur kopiert gar nichts
    * Wert der Funktion ist entry.
    *)
BEGIN
   RETURN entry;
END CopyNothing;

PROCEDURE CopyEntry(entry: ADDRESS): ADDRESS;
   (*
    * Kopiert den Speicherbereich auf den entry zeigt.
    *)
VAR   result:  ADDRESS;
BEGIN
   ALLOCATE(result, SIZE(entry^));
   Move(entry, result, SIZE(entry^));
   RETURN result;
END CopyEntry;


(* Standard-Proceduren fr LCondProc *)

PROCEDURE Ever(entry, info: ADDRESS): BOOLEAN;
   (* liefert IMMER TRUE *)
BEGIN
   RETURN TRUE;
END Ever;

PROCEDURE Equal(entry, parm: ADDRESS): BOOLEAN;
   (* testet entry und parm auf (Pointer-)Gleichheit *)
BEGIN
   RETURN (entry=parm);
END Equal;

BEGIN
   ModReg.Register('Lists',0,0,' ',1996,11,06,PROC(NIL));

   
END Lists.
